#include "jukebox.h"
#include "buttons.h"

enum cmodes continue_mode = cm_album;

#define BUFSIZE 1024
unsigned char buf[BUFSIZE];

int player_pid = 0;
int record_pid = 0;

int display_off_time = 0;
int timer_stop_time = 0;

int refresh_time = 0;

int pipe_fd[2];

char song_display[80];
char radio_display[80];
char findbuf[80];
int alphacnt;
#define ALPHA_LENGTH 30
char patbuf[ALPHA_LENGTH+1];
char prev_patbuf[ALPHA_LENGTH+1];

int find_mode = 0;
int radio_mode = 0;
int recording = 0;
int raw_record;
int rec_secs;
int found_song;

time_t rec_toggle_time = 0;
time_t raw_toggle_time = 0;

int album_show_mode = 0;

void sigchld_handler()
{
   int status;
   int pid;
   pid = waitpid(-1,&status,WNOHANG);
/*
   printf("waitpid returned: player_pid=%d, pid=%d, EXIT=%d, SIG=%d\n",
          player_pid, pid, WIFEXITED(status), WIFSIGNALED(status));
   if(WIFSIGNALED(status)) printf("termsig=%0d\n",WTERMSIG(status));
*/
   if(((player_pid && (pid == player_pid))
       || (record_pid && pid == record_pid))
      && (WIFEXITED(status) || WIFSIGNALED(status))) {
      char c = 0;
      //printf("pid %d exited\n",pid);
      write(pipe_fd[1],&c,1);
   }
   signal(SIGCHLD,&sigchld_handler);
}

int paused;

int profile = 0;
int profile_flag = 0;
int cuckoo_vol_flag = 0;
int muted_flag = 0;

enum {wn_stop,wn_cont,wn_play,wn_time,wn_speakfile,wn_cuckoomode,
      wn_cuckootest,wn_continuemode,wn_radio} what_next;

insert_str(char *dest, char *src, int n) {
   while(n--) {
      if(src && *src) *dest++ = *src++;
      else *dest++ = ' ';
   }
}

do_stop()
{
   if(player_pid) {
      if(paused) kill(player_pid,SIGCONT);
      kill(player_pid,SIGTERM);
   }
}

enum pl_mode {pl_song,pl_time,pl_speakfile,pl_cuckoo,pl_cuckoomode,pl_cuckootest,pl_continuemode} play_mode;
int play_what;
char *play_whatfile;

int volume = 70;
int saved_volume;

int starting_percent;

do_play(enum pl_mode what)
{
   //printf("do_play(%d)\n",what);
   play_mode = what;
   if(player_pid) {
      printf("Already playing.\n");
   } else {
      paused = 0;
      if(play_mode == pl_cuckoo) {
         saved_volume = volume;
         volume = cuckoo_volume;
         set_volume(0);
      }
      player_pid = fork();
      if(!player_pid) {
         switch(play_mode) {
            case pl_song:
                if(track_data[current_track].album == radio_record_album) {
                   play_recording(track_data[current_track].name);
                } else if(starting_percent) {
                   int starting_block = track_data[current_track].length;
                   char nbuf[20];
                   if(starting_block > 0) starting_block = (starting_percent * starting_block) / 100;
                   sprintf(nbuf,"%d",starting_block);
                   //printf("starting_block = %s\n",nbuf);
                   execl(MPG123,MPG123,MPG123_DEV "-q","-k",nbuf,
                         track_data[current_track].name,0);
                } else {
                   execl(MPG123,MPG123,MPG123_DEV "-q",
                         track_data[current_track].name,0);
                }
            case pl_time:
                say_time();
                _exit(0);
            case pl_speakfile:
               open_audio();
               speak_file(play_whatfile);
               close_audio();
               _exit(0);
            case pl_cuckoo:
               do_cuckoo(play_what);
               _exit(0);
            case pl_cuckoomode:
               do_cuckoomode();
               _exit(0);
            case pl_cuckootest:
               test_cuckoo();
               _exit(0);
            case pl_continuemode:
               say_continuemode();
               _exit(0);
         }
      }
      //printf("player launched on pid %d\n",player_pid);
      if(play_mode == pl_song) {
         hist_new(current_track);
         show_songinfo();
      }
      paused = 0;
      what_next = (play_mode == pl_song) ? wn_cont : wn_stop;
   }
}

do_restart()
{
   if(player_pid) {
      what_next = wn_play;
      do_stop();
   } else if(record_pid) {
      what_next = wn_play;
   } else {
      do_play(pl_song);
   }
}

set_volume(int delta)
{
   int mainvol;
   int pcmvol;

   volume += delta;
   if(volume<0) volume = 0;
   if(volume>115) volume=115;
   //printf("volume=%d\n",volume);

   mainvol = volume;
   pcmvol = 85;
   if(volume>100) {
      mainvol = 100;
      pcmvol = volume-15;
   }

   mixer_set(mix_mainvol,mainvol);
   mixer_set(mix_pcm,pcmvol);

   if(muted_flag) {
      muted_flag = 0;
      show_songinfo();
      show_radioinfo();
   }

}

int flash_state = 0;

update_songinfo()
{
   insert_str(song_display,track_data[current_track].title,40);
   if(track_data[current_track].fave_flags & (1<<profile)) song_display[39]='*';
   if(muted_flag && flash_state) insert_str(song_display+33," MUTE",5);
   if(album_show_mode) {
      insert_str(song_display+40,album_data[track_data[current_track].album].title,40);
   } else {
      insert_str(song_display+40,track_data[current_track].artist,40);
   }
   if(!player_pid) {
      insert_str(song_display+75," STOP",5);
   } else if(paused) {
      if(flash_state) insert_str(song_display+73," PAUSED",7);
   } else {
      insert_str(song_display+75," PLAY",5);
   }
}

enter_radio_mode()
{
   radio_mode = 1;
   recording = 0;
   display_off_time = 0;
   radio_unmute();
   mixer_set(mix_radio,85);
   show_radioinfo();
   record_pid = fork();
   if(!record_pid) {
      record_main();
   }
}

exit_radio_mode()
{
   if(radio_mode) {
      mixer_set(mix_radio,0);
      radio_mute();
      radio_mode = 0;
      refresh_display();
      schedule_display_off();
      find_recordings();
   }
   if(record_pid) {
      char c = REC_CTRL_QUIT;
      write(rec_ctrl_fd[1],&c,1);
   }
}

update_radio_display()
{
   char buf[80];
   int i;
   char *n;
   sprintf(buf,"FM %d.%d",radio_freq/10,radio_freq%10);
   insert_str(radio_display,buf,40);
   n = presetname(radio_freq);
   if(n) insert_str(radio_display+40,n,40);
   else insert_str(radio_display+40," ",40);
   if(muted_flag && flash_state) insert_str(radio_display+33," MUTE",5);
   if(recording) {
       int offset = 67;
       int length = 13;
       char *rawstr = "";
       char rsbuf[30];
       if(raw_record) {
          offset = 63;
          rawstr = "RAW ";
          length = 17;
       } 
       sprintf(rsbuf," %s%02d:%02d:%02d REC",rawstr,rec_secs/3600,(rec_secs/60)%60,rec_secs%60);
       insert_str(radio_display+offset,rsbuf,length);
   }
}

show_songinfo()
{
   update_songinfo();
   refresh_display();
}

show_radioinfo()
{
   update_radio_display();
   refresh_display();
}

refresh_display()
{
   if(radio_mode) {
      update_lcd(radio_display);
      update_all_remote_displays(radio_display);
   } else if(find_mode) {
      display_find_result();
   } else {
      update_lcd(song_display);
      update_all_remote_displays(song_display);
   }
}

int radio_unmute_flag = 0;

radio_tune_with_mute()
{
    show_radioinfo();
    radio_mute();
    radio_tune(radio_freq);
    radio_unmute_flag = 1;
}

show_temp_message(int timeout,int clear,int pos,char *message)
{
   char tmbuf[80];
   time_t tt = time(0);
   int i;
   for(i=0;i<80;i++) tmbuf[i] = clear ? ' ' : lcd_contents[i];
   i = 0;
   while(message[i] && pos<80) {
      tmbuf[pos++] = message[i++];
   }
   update_lcd(tmbuf);
   update_all_remote_displays(tmbuf);
   if(timeout) refresh_time = tt + timeout;
}

show_volume(char *what,int level)
{
   char svbuf[41];
   int i;
   insert_str(svbuf,what,40);
   for(i=0;i<level;i+=5) svbuf[15+i/5] = '#';
   show_temp_message(3,0,40,svbuf);
}

display_find_result()
{
   sprintf(findbuf+40,"[%d]    ",find_count);
   insert_str(findbuf+47,find_count?track_data[found_song].title:"No match",33);
   update_lcd(findbuf);
   update_all_remote_displays(findbuf,-1);
}

schedule_display_off()
{
   time_t tt = time(0);
   display_off_time = tt+60;
}


main()
{
   fd_set readfds;
   int code = 0;
   int irfd;
   FILE *fp;
   int repeat_code = 0;
   int saved_code = -1;

   prev_patbuf[0] = 0;

   init_history();

   init_lcd();
   radio_init();
   init_cuckoo();

   signal(SIGCHLD,&sigchld_handler);

   srandom(time(0));

   init_songdb();

   if(pipe(pipe_fd)<0 || pipe(rec_ctrl_fd)<0 || pipe(rec_status_fd)<0) {
      perror("pipe");
      exit(999);
   } 

   irfd = open("/dev/ttyS0",0);
   if(irfd<0) {
      perror("serial open");
      exit(999);
   }

   set_volume(0);

   mixer_set(mix_radio,0);

   show_songinfo();
   schedule_display_off();

   init_remoteif();

   init_timer();

   for(;;) {
      struct timeval tv, *tvp;
      int r;
      int sel_result;
      int maxfd = irfd;
      time_t tt = time(0);
      int i;
      int remote_code;
      if(pipe_fd[0]>irfd) maxfd=pipe_fd[0];
      if(rec_status_fd[0]>maxfd) maxfd=rec_status_fd[0];
      FD_ZERO(&readfds);
      FD_SET(irfd,&readfds);
      FD_SET(pipe_fd[0],&readfds);
      FD_SET(rec_status_fd[0],&readfds);

      remoteif_do_fd_set(&readfds,&maxfd);

      if(player_pid) display_off_time = 0;

      tvp = 0;
      if(radio_unmute_flag) {
         tv.tv_sec = 0;
         tv.tv_usec = 100000;
         tvp = &tv;
      } else if(repeat_code || paused || muted_flag || rec_toggle_time || raw_toggle_time) {
         tv.tv_sec = 0;
         tv.tv_usec = 500000;
         tvp = &tv;
      } else {
         unsigned long wakeup_time = 0xffffffff;
         if(timer_wakeup_time) wakeup_time = timer_wakeup_time;
         if(cuckoo_mode != ck_off && next_cuckoo_time < wakeup_time) wakeup_time = next_cuckoo_time;
         if(timer_stop_time && timer_stop_time < wakeup_time) wakeup_time = timer_stop_time;
         if(refresh_time && refresh_time < wakeup_time) wakeup_time = refresh_time;
         if(display_off_time && display_off_time < wakeup_time) wakeup_time = display_off_time;
         if(wakeup_time != 0xffffffff) {
            int delay = (unsigned long) wakeup_time - tt;
            if(delay<0) delay = 0;
            tv.tv_usec = 0;
            tv.tv_sec = delay;
            tvp = &tv;
         }
      }
/*
      if(tvp) {
         printf("calling select with delay %d.%d\n",tvp->tv_sec,tvp->tv_usec);
      } else {
         printf("calling select with no timeout\n");
      }
*/
      do {
         sel_result=select(maxfd+1,&readfds,0,0,tvp);
      } while(sel_result<0 && errno==EINTR);
      if(sel_result<0) {
         perror("select");
         exit(999);
      }

      remote_code = 0;

      if(sel_result) {
         remote_code = remoteif_do_fd_check(&readfds);
      }

      // Hack to get rid of the timeout when record is toggled via
      // the telnet interface

      if(radio_mode && remote_code == ALBUM_TOGGLE_CODE) {
         remote_code = 0;
         rec_toggle_time = time(0);
      }
      if(radio_mode && remote_code == PROFILE_CODE) {
         remote_code = 0;
         raw_toggle_time = time(0);
      }

      if(!sel_result && radio_unmute_flag) {
         radio_unmute_flag = 0;
         radio_unmute();
      }

      if(!sel_result && (paused || muted_flag)) {
         show_songinfo();
         show_radioinfo();
         flash_state = !flash_state;
      }
      tt = time(0);

      if(rec_toggle_time && tt >= rec_toggle_time) {
         rec_toggle_time = 0;
         if(recording) {
            char c = REC_CTRL_STOP;
            write(rec_ctrl_fd[1],&c,1);
         } else {
            char c = REC_CTRL_START;
            write(rec_ctrl_fd[1],&c,1);
            timer_stop_time = 0;
            write(rec_ctrl_fd[1],&c,1);
         }
      }

      if(raw_toggle_time && tt >= raw_toggle_time) {
         char c = REC_CTRL_DONT_COMPRESS;
         raw_toggle_time = 0;
         write(rec_ctrl_fd[1],&c,1);
      }

      if(!sel_result && display_off_time && tt >= display_off_time) {
         if(!lcd_is_off) {
            lcd_off();
            update_all_remote_displays(lcd_contents);
         }
         if(find_mode && *patbuf) strcpy(prev_patbuf,patbuf);
         find_mode = 0;
         display_off_time = 0;
      }

      if(next_cuckoo_time && tt >= next_cuckoo_time) {
         if(cuckoo_mode != ck_off) {
            struct tm *t = localtime(&tt);
            if(!player_pid) {
               if(radio_mode) {
                  printf("cuckoo:  no strike because radio is on\n");
               } else if(!(cuckoo_mode==ck_daytime && (t->tm_hour > 22 || t->tm_hour < 7))) {
                  //play_what = t->tm_min % 24;
                  play_what = t->tm_hour;
                  printf("cuckoo: hour=%d\n",play_what);
                  do_play(pl_cuckoo);
               } else {
                  printf("daytime cuckoo:  no strike because hour=%d\n",t->tm_hour);
               }
            }
            schedule_next_cuckoo();
         } else {
            next_cuckoo_time = 0;
         }
      }
      if(refresh_time && tt >= refresh_time) {
         refresh_time = 0;
         refresh_display();
      }
      if(timer_wakeup_time && tt >= timer_wakeup_time) {
         if(recording) {
            // if already recording, forget the timer event
            printf("timer event cancelled because record in progress\n");
         } else {
            char c;
            if(!radio_mode) {
               if(player_pid) {
                  // something is playing, kill it
                  what_next = wn_radio;
                  do_stop();
               } else {
                  int my_saved_volume = volume;
                  // The system was idle, so to avoid surprising anyone
                  // when recording starts up, set volume to zero.
                  set_volume(-200);
                  enter_radio_mode();
                  volume = my_saved_volume;
               }
            }
            radio_freq = timer_freq;
            radio_tune_with_mute();
            printf("timer record start\n");
            timer_stop_time = timer_wakeup_time + timer_duration * 60;

            // it should be possible to send the record start command even
            // if the record task is not running yet -- it will just read
            // it when it's ready.

            c = REC_CTRL_START;
            write(rec_ctrl_fd[1],&c,1);
         }
         schedule_next_timer_event();
      }

      if(timer_stop_time && tt >= timer_stop_time) {
         timer_stop_time = 0;
         if(recording) {
            char c = REC_CTRL_STOP;
            write(rec_ctrl_fd[1],&c,1);
            printf("timer record stop\n");
            exit_radio_mode();
            what_next = wn_stop;
            // ensure the volume control register is up to date
            // (this undoes the mute at record start time if any)
            set_volume(0);
         }
      }

      if(FD_ISSET(rec_status_fd[0],&readfds)) {
         char c;
         read(rec_status_fd[0],&c,1);
         //printf("Got %d from record process\n",c);
         switch(c) {
            case REC_STATUS_RECORDING: {
               unsigned char c[2];
               if(!recording) raw_record = 0;
               recording = 1;
               read(rec_status_fd[0],&c,2);
               rec_secs = c[0] | (c[1] << 8);
               show_radioinfo();
               break;
            }
            case REC_STATUS_OK_NO_COMPRESS:
               raw_record = 1;
               show_radioinfo();
               break;
            case REC_STATUS_STOPPED:
               recording = 0;
               show_radioinfo();
               break;
            case REC_STATUS_DISK_FULL:
               if(recording) {
                  insert_str(radio_display+40,"*** RECORD STOPPED: OUT OF SPACE ***",40);
               } else {
                  insert_str(radio_display+40,"*** CAN'T RECORD: OUT OF SPACE ***",40);
               }
               recording = 0;
               refresh_display();
               break;
         }
      }

      if(FD_ISSET(pipe_fd[0],&readfds)) {
         char c;
         r = read(pipe_fd[0],&c,1);
         //printf("subprocess exited.\n");
         schedule_display_off();
         paused = 0;
         player_pid = 0;
         record_pid = 0;
         if(recording) {
            recording = 0;
            show_radioinfo();
         }
         
         if(play_mode == pl_cuckoo) {
            volume = saved_volume;
            set_volume(0);
         }
         if(play_mode == pl_song && what_next != wn_cont && what_next != wn_play) show_songinfo();
         switch(what_next) {
            case wn_cont:
               switch(continue_mode) {
                  case cm_single:
                     show_songinfo();
                     break;
                  case cm_trackloop:
                     starting_percent = 0;
                     do_play(pl_song);
                     break;
                  case cm_album: {
                        int alb = track_data[current_track].album;
                        int c = (current_track+1) % tracks;
                        if(track_data[c].album == alb) {
                           current_track = c;
                           starting_percent = 0;
                           do_play(pl_song);
                        } else show_songinfo();
                     }
                     break;
                  case cm_albumloop: {
                        int alb = track_data[current_track].album;
                        do {
                           current_track = (current_track+1) % tracks;
                        } while (track_data[current_track].album != alb);
                        starting_percent = 0;
                        do_play(pl_song);
                     }
                     break;
                  case cm_continuous:
                     current_track++;
                     if(current_track==tracks) current_track = 0;
                     starting_percent = 0;
                     do_play(pl_song);
                     break;
                  case cm_faveloop: {
                        int here = current_track;
                        do {
                           current_track = (current_track+1) % tracks;
                        } while(current_track != here && !(track_data[current_track].fave_flags&(1<<profile)));
                        if(track_data[current_track].fave_flags&(1<<profile)) {
                           starting_percent = 0;
                           do_play(pl_song);
                        }
                     }
                     break;
               }
               break;
            case wn_play:
               do_play(pl_song);
               break;
            case wn_time:
               do_play(pl_time);
               break;
            case wn_speakfile:
               do_play(pl_speakfile);
               break;
            case wn_cuckoomode:
               do_play(pl_cuckoomode);
               break;
            case wn_cuckootest:
               do_play(pl_cuckootest);
               break;
            case wn_continuemode:
               do_play(pl_continuemode);
               break;
            case wn_radio:
               enter_radio_mode();
               break;
         }
      }
      if((!sel_result && repeat_code) || remote_code || FD_ISSET(irfd,&readfds)) {
         char c;
         int got_rc_code = 0;
         if(remote_code) {
            code = remote_code;
         } else if(!sel_result) {
            code = repeat_code;
            got_rc_code = 1;
         } else {
            r = read(irfd,&c,1);
            if(r!=1) {
               fprintf(stderr,"read returned %d\n",r);
               exit(999);
            }
             if(c=='\r' || c=='\n') got_rc_code = 1;
            //printf("got %c\n",c);
         }
         repeat_code = 0;
         if(!sel_result || remote_code || got_rc_code) {
            if(lcd_is_off) refresh_display();
            if(!radio_mode) schedule_display_off();
            //printf("in there, sel_result = %d\n",sel_result);
            //if(code) printf("code: %04x\n",code); if(0) 
            if(find_mode) {
               int i;
               char c = 0;
               int upd = 0;
               for(i=0;i<26;i++) {
                  if(code==alpha_codes[i]) {
                     c = i+'A';
                     break;
                  }
               }
               if(code == STAR_CODE) c = '*';
               if(c) {
                  if(alphacnt < ALPHA_LENGTH) {
                     patbuf[alphacnt] = c;
                     findbuf[6+alphacnt++] = c;
                     findbuf[6+alphacnt] = ' ';
                     patbuf[alphacnt] = 0;
                     found_song = find_tracks(patbuf);
                     upd = 1;
                  }
               } else if(code==POUND_CODE) {
                  if(!alphacnt) {
                     strcpy(patbuf,prev_patbuf);
                     insert_str(findbuf+6,patbuf,30);
                     alphacnt = strlen(patbuf);
                     found_song = find_tracks(patbuf);
                     display_find_result();
                     upd = 0;
                  }
               } else if(code==LEFT_CODE) {
                  if(alphacnt) {
                     findbuf[6+--alphacnt] = ' ';
                     patbuf[alphacnt] = 0;
                     found_song = -1;
                     found_song = find_tracks(patbuf);
                     upd = 1;
                  }
               } else if(code==STOP_CODE) {
                  find_mode = 0;
                  if(*patbuf) strcpy(prev_patbuf,patbuf);
                  show_songinfo();
               } else if(code==UP_CODE) {
                  int here = found_song;
                  if(find_count > 0) {
                     do {
                        found_song = found_song-1;
                        if(found_song<0) found_song = tracks-1;
                     } while(found_song != here && !track_data[found_song].found);
                     upd = 1;
                  }
               } else if(code==DOWN_CODE) {
                  int here = found_song;
                  if(find_count > 0) {
                     do {
                        found_song = (found_song+1) % tracks;
                     } while(found_song != here && !track_data[found_song].found);
                     upd = 1;
                  }
               } else if(code==PLAY_CODE) {
                  if(find_count > 0) {
                     exit_radio_mode();
                     find_mode = 0;
                     if(*patbuf) strcpy(prev_patbuf,patbuf);
                     current_track = found_song;
                     starting_percent = 0;
                     do_restart();
                     upd = 0;
                  }
               } else if(code==RIGHT_CODE) {
                  if(find_count > 0) {
                     find_mode = 0;
                     if(*patbuf) strcpy(prev_patbuf,patbuf);
                     exit_radio_mode();
                     current_track = found_song;
                     if(player_pid) {
                        what_next = wn_stop;
                        do_stop();
                     } else if(record_pid) {
                        what_next = wn_stop;
                     } else {
                        show_songinfo();
                     }
                     upd = 0;
                  }
               } else if(code==0) {
                  // key up.
                  if(!saved_code) {
                     init_lcd();
                     refresh_display();
                     code = -1;         // hack
                  }
               } else {
                  printf("code: %04x\n",code);
               }
               if(upd) {
                  display_find_result();
                  if(find_count==1) {
                     exit_radio_mode();
                     if(*patbuf) strcpy(prev_patbuf,patbuf);
                     find_mode = 0;
                     exit_radio_mode();
                     current_track = found_song;
                     starting_percent = 0;
                     do_restart();
                  }
               }  
            } else switch(code) {
               int digit;
               case ZERO_CODE:
                  digit = 0;
                  goto numeric_code;
               case ONE_CODE:
                  digit = 1;
                  goto numeric_code;
               case TWO_CODE:
                  digit = 2;
                  goto numeric_code;
               case THREE_CODE:
                  digit = 3;
                  goto numeric_code;
               case FOUR_CODE:
                  digit = 4;
                  goto numeric_code;
               case FIVE_CODE:
                  digit = 5;
                  goto numeric_code;
               case SIX_CODE:
                  digit = 6;
                  goto numeric_code;
               case SEVEN_CODE:
                  digit = 7;
                  goto numeric_code;
               case EIGHT_CODE:
                  digit = 8;
                  goto numeric_code;
               case NINE_CODE:
                  digit = 9;
numeric_code:
                  if(recording) break;
                  if(radio_mode) {
                     if(digit==0) digit=9; else digit--;
                     if(set_radio_preset(digit)) {
                        radio_tune_with_mute();
                     }
                  } else if(profile_flag) {
                     char pmbuf[40];
                     play_whatfile = numfilename(digit);
                     profile = digit;
                     show_songinfo();
                     if(!player_pid) {
                        do_play(pl_speakfile);
                     } else {
                        what_next = wn_speakfile;
                        do_stop();
                     }
                     sprintf(pmbuf,"Profile %d selected",profile);
                     show_temp_message(3,1,0,pmbuf);
                  } else {
                     int l = track_data[current_track].length;
                     if(l>0) {
                        starting_percent = digit*10;
                        do_restart();
                     }
                  }
                  break;
               case PLAY_CODE:
                  if(recording) break;
                  exit_radio_mode();
                  if(record_pid) {
                     //printf("play: record_pid=%d\n",record_pid);
                     starting_percent = 0;
                     what_next = wn_play;
                  } else if(!player_pid) {
                     //printf("play: player_pid=0, starting up\n");
                     starting_percent = 0;
                     do_play(pl_song);
                  } else {
                     //printf("play: pause/unpause\n");
                     if(play_mode==pl_song) {
                        //printf("kill -%d %d\n",paused ? SIGCONT : SIGSTOP,player_pid);
                        kill(player_pid,paused ? SIGCONT : SIGSTOP);
                        paused = !paused;
                        show_songinfo();
                     } else {
                        starting_percent = 0;
                        what_next = wn_play;
                        do_stop();
                     }
                  }
                  break;
               case STOP_CODE:
                  if(recording) break;
                  if(radio_mode) {
                     exit_radio_mode();
                     what_next = wn_stop;
                  } else if(record_pid) {
                     what_next = wn_stop;
                  } else if(player_pid && play_mode == pl_song) {
                     what_next = wn_stop;
                     do_stop();
                  } else {
                     switch(continue_mode) {
                        case cm_single: {
                              continue_mode = cm_trackloop;
                              show_temp_message(3,1,0,"Track Loop Mode");
                           }
                           break;
                        case cm_trackloop: {
                              continue_mode = cm_album;
                              show_temp_message(3,1,0,"Single Album Mode");
                           }
                           break;
                        case cm_album: {
                              continue_mode = cm_albumloop;
                              show_temp_message(3,1,0,"Album Loop Mode");
                           }
                           break;
                        case cm_albumloop: {
                              continue_mode = cm_continuous;
                              show_temp_message(3,1,0,"Infinite Play Mode");
                           }
                           break;
                        case cm_continuous: {
                              continue_mode = cm_faveloop;
                              show_temp_message(3,1,0,"Favourites Loop Mode");
                           }
                           break;
                        case cm_faveloop: {
                              continue_mode = cm_single;
                              show_temp_message(3,1,0,"Single Track Mode");
                           }
                           break;
                     }
                     if(player_pid) {
                        what_next = wn_continuemode;
                        do_stop();
                     } else {
                        do_play(pl_continuemode);
                     }
                  }
                  break;
               case UP_CODE:
                  if(recording) break;
                  if(radio_mode) {
                     find_next_radio_preset(1);
                     radio_tune_with_mute();
                  } else {
                     int a;
                     a = track_data[current_track].album - 1;
                     if(a<0) a = albums-1;
                     current_track = album_data[a].first_track;
                     starting_percent = 0;
                     if(player_pid && play_mode == pl_song) do_restart(); else {
                        update_songinfo();
                        show_temp_message(5,1,0,album_data[a].title);
                        if(album_data[a].artist) show_temp_message(5,0,40,album_data[a].artist);
                     }
                  }
                  break;                  
               case DOWN_CODE:
                  if(recording) break;
                  if(radio_mode) {
                     find_next_radio_preset(-1);
                     radio_tune_with_mute();
                  } else {
                     int here = current_track;
                     int a;
                     do {
                        current_track++;
                        if(current_track>=tracks) current_track = 0;
                        if((a=track_data[current_track].album) != track_data[here].album) break;
                        
                     } while(current_track != here);
                     starting_percent = 0;
                     if(player_pid && play_mode == pl_song) do_restart(); else {
                        update_songinfo();
                        show_temp_message(5,1,0,album_data[a].title);
                        if(album_data[a].artist) show_temp_message(5,0,40,album_data[a].artist);
                     }
                  }
                  break;                  
               case RIGHT_CODE: 
                  if(recording) break;
                  if(radio_mode) {
                     radio_tune_up();
                     radio_tune_with_mute();
                  } else { 
                     int a = track_data[current_track].album;
                     current_track++;
                     if(track_data[current_track].album != a) current_track = album_data[a].first_track;
                     starting_percent = 0;
                     if(player_pid && play_mode == pl_song) do_restart(); else show_songinfo();
                  }
                  break;
               case HIST_BACKWARD_CODE: {
                     int n;
                     if(recording) break;
                     n = hist_backward();
                     exit_radio_mode();
                     if(n>=0) {
                        current_track = n;
                        starting_percent = 0;
                        if(record_pid ||(player_pid && play_mode == pl_song)) do_restart(); else show_songinfo();
                     }
                  }
                  break;
               case HIST_FORWARD_CODE: {
                     int n;
                     if(recording) break;
                     n = hist_forward();
                     exit_radio_mode();
                     if(n>=0) {
                        current_track = n;
                        starting_percent = 0;
                        if(record_pid ||(player_pid && play_mode == pl_song)) do_restart(); else show_songinfo();
                     }
                  }
                  break;
               case LEFT_CODE:
                  if(recording) break;
                  if(radio_mode) {
                     radio_tune_down();
                     radio_tune_with_mute();
                  } else { 
                     int a = track_data[current_track].album;
                     exit_radio_mode();
                     current_track--;
                     if(current_track<0) {
                        if(albums==1) current_track = tracks-1;
                        else current_track = album_data[1].first_track-1;
                     } else if(track_data[current_track].album != a) {
                        if(a == (albums-1)) current_track = tracks-1;
                        else current_track = album_data[a+1].first_track-1;
                     }
                     starting_percent = 0;
                     if(record_pid || (player_pid && play_mode == pl_song)) do_restart(); else show_songinfo();
                  }
                  break;
               case FAVE_TOGGLE_CODE:
                  if(track_data[current_track].album != radio_record_album) {
                     char fmbuf[42];
                     FILE *fp;
                     if(recording) break;
                     exit_radio_mode();
                     if(record_pid) what_next = wn_stop;
                     track_data[current_track].fave_flags ^= (1<<profile);
                     if(track_data[current_track].fave_flags&(1<<profile)) {
                        song_display[39] = '*';
                        sprintf(fmbuf,"%-41s","*Added to favourites");
                     } else {
                        song_display[39] = ' ';
                        sprintf(fmbuf,"%-41s"," Removed from favourites");
                     }
                     show_temp_message(3,0,39,fmbuf);
                     fp = fopen("favourites","a");
                     fprintf(fp,"%03x %s\n",track_data[current_track].fave_flags,track_data[current_track].name);
                     fclose(fp);
                  }
                  break;
               case NEXT_FAVE_CODE: {
                     int here = current_track;
                     if(recording) break;
                     exit_radio_mode();
                     what_next = wn_stop;
                     do {
                        current_track = (current_track+1) % tracks;
                     } while(current_track != here && !(track_data[current_track].fave_flags&(1<<profile)));
                     if(current_track==here &&  !(track_data[current_track].fave_flags&(1<<profile))) {
                        //if(!player_pid) {
                        //   play_whatfile = "no_favourites";
                        //   do_play(pl_speakfile);
                        //}
                        show_temp_message(3,1,0,"Favourite list is empty.");
                        break;
                     }
                     starting_percent = 0;
                     if(record_pid || (player_pid && play_mode == pl_song && current_track != here)) do_restart();
                     else show_songinfo();
                  }
                  break;
               case PREV_FAVE_CODE: {
                     int here = current_track;
                     if(recording) break;
                     exit_radio_mode();
                     what_next = wn_stop;
                     do {
                        current_track--;
                        if(current_track < 0) current_track = tracks-1;
                     } while(current_track != here && !(track_data[current_track].fave_flags&(1<<profile)));
                     if(current_track==here &&  !(track_data[current_track].fave_flags&(1<<profile))) {
                        //if(!player_pid) {
                        //   play_whatfile = "no_favourites";
                        //   do_play(pl_speakfile);
                        //}
                        show_temp_message(3,1,0,"Favourite list is empty.");
                        break;
                     }
                     starting_percent = 0;
                     if(record_pid || (player_pid && play_mode == pl_song && current_track != here)) do_restart();
                     else show_songinfo();
                  }
                  break;
               case VOL_UP_CODE:
                  if(cuckoo_vol_flag) {
                     cuckoo_volume += 5;
                     if(cuckoo_volume > 115) cuckoo_volume = 115;
                     cuckoo_vol_flag = 2;
                     if(player_pid) {
                        what_next = wn_cuckootest;
                        do_stop();
                     } else {
                        do_play(pl_cuckootest);
                     }
                     show_volume("Cuckoo Volume",cuckoo_volume);
                  } else {
                     set_volume(5);
                     show_volume("Volume",volume);
                  }
                  if(got_rc_code) repeat_code = code;
                  break;
               case VOL_DOWN_CODE:
                  if(cuckoo_vol_flag) {
                     cuckoo_volume -= 5;
                     if(cuckoo_volume < 0) cuckoo_volume = 0;
                     cuckoo_vol_flag = 2;
                     if(player_pid) {
                        what_next = wn_cuckootest;
                        do_stop();
                     } else {
                        do_play(pl_cuckootest);
                     }
                     show_volume("Cuckoo Volume",cuckoo_volume);
                  } else {
                     set_volume(-5);
                     show_volume("Volume",volume);
                  }
                  if(got_rc_code) repeat_code = code;
                  break;
               case MUTE_CODE:
                  if(muted_flag) {
                     set_volume(0);
                  } else {
                     int my_saved_volume = volume;
                     set_volume(-30);
                     muted_flag = 1;
                     volume = my_saved_volume;
                  }
                  flash_state = 1;
                  show_songinfo();
                  show_radioinfo();
                  break;
               case TIME_CODE:
                  if(recording) break;
                  exit_radio_mode();
                  if(record_pid) {
                     what_next = wn_time;
                  } else if(player_pid) {
                     what_next = wn_time;
                     do_stop();
                  } else {
                     do_play(pl_time);
                  }
                  break;  
               case RADIO_CODE:
                  if(recording) break;
                  if(radio_mode) {
                     exit_radio_mode();
                     what_next = wn_stop;
                  } else if(have_radio) {
                     if(player_pid) {
                        what_next = wn_radio;
                        do_stop();
                     } else {
                        enter_radio_mode();
                     }
                  }
                  break;
               case POUND_CODE: {
                     time_t t = time(0);
                     char *ct = ctime(&t);
                     FILE *fp;
                     if(radio_mode) break;
                     fp = fopen("poundkey.log","a");
                     while(*ct && *ct != '\n') putc(*ct++,fp); 
                     fprintf(fp,"  %s\n",track_data[current_track].name);
                     fclose(fp);
                  } 
                  break;
               case STAR_CODE:
                  if(!recording && tracks > real_tracks) {
                     what_next = wn_stop;
                     if(player_pid) do_stop();
                     if(radio_mode) exit_radio_mode();
                     if(current_track <= real_tracks) current_track = real_tracks;
                     show_songinfo();
                  }
                  break;
               case PROFILE_CODE:
                  if(radio_mode) {
                     raw_toggle_time = time(0) + 2;
                  } else {
                     play_whatfile = "profile";
                     if(record_pid) {
                        what_next = wn_speakfile;
                     } else if(player_pid) {
                        what_next = wn_speakfile;
                        do_stop();
                     } else {
                        do_play(pl_speakfile);
                     }
                     profile_flag = 2;
                  }
                  break;
               case RANDOM_CODE:
                  if(recording) break;
                  exit_radio_mode();
                  current_track = random() % tracks;
                  if(record_pid) {
                     starting_percent = 0;
                     what_next = wn_play;
                  } else if(!player_pid) {
                     starting_percent = 0;
                     do_play(pl_song);
                  } else {
                     starting_percent = 0;
                     what_next = wn_play;
                     do_stop();
                  }
                  break;
               case CUCKOO_CODE: {
                     char *message;
                     if(recording) break;
                     exit_radio_mode();
                     switch(cuckoo_mode) {
                        case ck_off:
                           cuckoo_mode = ck_daytime;
                           schedule_next_cuckoo();
                           cuckoo_vol_flag = 2;
                           message = "Daytime Cuckoo Clock";
                           break;
                        case ck_daytime:
                           cuckoo_mode = ck_on;
                           schedule_next_cuckoo();
                           cuckoo_vol_flag = 2;
                           message = "24-Hour Cuckoo Clock";
                           break;
                        case ck_on:
                           cuckoo_mode = ck_off;
                           message = "No Cuckoo Clock";
                           break;
                     }
                     show_temp_message(3,1,0,message);
                     show_volume("Cuckoo Volume",cuckoo_volume);
                     if(record_pid) {
                        what_next = wn_cuckoomode;
                     } else if(player_pid) {
                        what_next = wn_cuckoomode;
                        do_stop();
                     } else {
                        do_play(pl_cuckoomode);
                     }
                  }
                  break;
               case FIND_CODE: {
                     int i;
                     if(recording) break;
                     strcpy(findbuf,"Find:");
                     for(i=5;i<80;i++) findbuf[i] = ' ';
                     find_mode = 1;
                     alphacnt = 0;
                     patbuf[0] = 0;
                     found_song = find_tracks(patbuf);
                     display_find_result();
                     show_temp_message(0,0,40,"Press # to recall last search pattern   ");
                  }
                  break;
               case ALBUM_TOGGLE_CODE:
                  if(radio_mode) {
                     rec_toggle_time = time(0) + 2;
                  } else {
                     album_show_mode = !album_show_mode;
                     show_songinfo();
                  }
                  break;
               case 0:
                  // button up
                  if(profile_flag) profile_flag++;  // ugly hack
                  if(cuckoo_vol_flag) cuckoo_vol_flag++;  // ugly hack

                  // if two button-ups in a row, initialize the LCD
                  if(!saved_code) {
                     init_lcd();
                     refresh_display();
                     code = -1;         // hack
                  }
                  rec_toggle_time = 0;
                  break;
               default:
                  printf("code: %04x\n",code);
            }
            if(code!=ALBUM_TOGGLE_CODE) rec_toggle_time = 0;
            saved_code = code;
            code = 0;
            if(cuckoo_vol_flag) cuckoo_vol_flag--;
            if(profile_flag) profile_flag--;
         } else if(c>='A' && c<='F') {
            code = 16*code + c - 'A' + 10;
         } else if(c>='a' && c<='f') {
            code = 16*code + c - 'A' + 10;
         } else if(c>='0' && c<='9') {
            code = 16*code + c - '0';
         }
      }
   }
}
