                .equ    keycode,0x28
		.equ	pausecount,0x29
                .equ	code,0x3a
		.flag	cdflag,0x20.0

		.org 0x0

                ; CPU initialization
                ; ------------------

                ; Initialize CPU registers.

start:          mov     SP,#0x2f        ; Stack at 30H
                mov     PSW,#0          ; Registers at 0H
                mov     TCON,#0         ; Disable timers
                mov     IE,#0           ; Disable all interrupts
                mov     IP,#0
                mov     PCON,#0         ; Initialize other registers.
                mov     P1,#0xff
                mov     P3,#0xff

                ; Serial port initialization
                ; --------------------------

                ; Set both timers to free running, 8 bit auto reload.

                mov     TMOD,#0x22

                ; Start baud rate generator and initialize serial port.

                mov     TH1,#0xfd       ; 9600 bps
                setb    TCON.6
                mov     SCON,#0x52

                clr     P3.3            ; R/W signal to LCD

                ; Send a bunch of key-up signals in a row to alert the
                ; jukebox program that the display pod has been reset.

                acall	senddot
		acall	senddot
		acall	senddot

		mov	code,#0
                sjmp	key0

		; Enter here to signal "key up"

sigkeyup:	mov	a,keycode
		jz	gloop
		acall	senddot

key0:           mov     keycode,#0

                ; wait for carrier

gloop:          acall   dolcd
                jb      P3.5,gloop

		; OK, so we got a carrier.  Time it to see if it's a
		; preamble.  Preambles are long, over 8000 instruction
		; cycles.

                mov	R0,#0

		; 30 * 256 = 7680, close enough.  So we must complete
		; 256 iterations of the following loop, which takes
		; 30 instruction cycles.

gcloop:		jb	P3.5,gotpause				; 2  ..2
                acall	dolcd                                   ; 26 ..28
		djnz	R0,gcloop				; 2  ..30

gotpreamble:	acall	waitpause
                mov	a,R0
                jz	sigkeyup

		; Then time the pause following.

gpa_ok:		acall	timepause

		; If it's a really long pause, signal key up and get out.

		jnz	sigkeyup

		; No pause that is part of a code has the high byte of its
		; time set, so if this is the case, treat it as a garbage
		; pause.

gp_notbail:	mov	a,R2
		jnz	gp_notkeyup

		; Pauses between 0x48 and 0x80 ticks directly following
		; a preamble mean "key still down".  Pauses shorter than
		; 0x48 are garbage.

		mov	a,R1
		clr	C
		subb	a,#0x48
		jc	gp_notkeyup

		; Check if it's a "key still down" code

		subb	a,#0x38
		jc	continue

		; Finally, continue here if we saw a valid preamble, then
		; a pause between 0x80 and 0x100 cycles long.  We are
		; currently in the pulse that separates pauses.  From
		; here on, we either signal a "key down" event or signal
		; "key up" and get out to "gloop".

                ; Grab 4 bytes

newkeydown:	acall	decodebyte
		jnz	sigkeyup
                mov     R7,3
		acall	decodebyte
                jnz	sigkeyup
                mov	R6,3
		acall	decodebyte
		jnz	sigkeyup
		mov	R5,3
		acall	decodebyte
		jnz	sigkeyup

                ; The second and 4th bytes are the complements of the
                ; ones preceding, check this.

                mov	a,R3
                cpl	a
                cjne	a,5,sigkeyup
                mov	a,R7
		cpl	a
		cjne	a,6,sigkeyup

                ; The key code is good.  Print it.

		mov	keycode,R5

                mov	a,R7
		acall	printhex2
		mov	a,R5
		acall	printhex2
		mov	a,#13
		acall	putchar

continue:	mov	pausecount,#10

		; Enter here if the IR code pulse we just saw wasn't a
		; preamble. 

		; Time what's left of the pause.  If greater than
		; the delay between "key still down" codes, signal a
		; "key up" event.

gotpause:	acall	timepause

		; Whenever we jump to the following, A nonzero means an
		; unacceptably long pause time which means key up.

           	jnz	sigkeyup

		; Be sure to signal a "key up" event if we come
		; through here too many times without anything making
		; sense.  This prevents IR noise from causing a bogus
		; "key still down" indication.

gp_notkeyup:	djnz	pausecount,gloop
		sjmp	sigkeyup

		; Subroutine to decode a byte of IR remote control
		; data.  Returns decoded byte in R3, A set nonzero if
		; anything went wrong and R3 is invalid.

decodebyte:	mov	R3,#0
		mov	R4,#8
decodeloop:	acall	waitpause
		mov	a,R0
		jz	db_timeout
		acall	timepause
		jnz	db_timeout
		mov	a,R2
		jnz	db_timeout
		mov	a,R1
		clr	C
		subb	a,#0x28
		mov	a,R3
		rlc	a
		mov	R3,a
		mov	a,R1
		clr	C
		subb	a,#0x50
		jnc	db_timeout
		djnz	R4,decodeloop
		clr	a
db_ret:		ret

db_timeout:	mov	a,#1
		ret

		; Subroutine to wait out the end of a pulse.  Returns
		; R0 set to zero if a timeout occurs.  The length of
                ; the timeout is not critical, it's just there to
                ; prevent failing on account of a "stuck on" IR source.

waitpause:	mov	R0,#255
wploop:   	jb	P3.5,wp_ret
                acall	dolcd
		mov	R1,#10
wp_delay:	djnz	R1,wp_delay
		djnz	R0,wploop
wp_ret:		ret

		; And here's the all-important subroutine to time the
		; length of pauses.  Works on a 46-instruction cycle.

		; Start the count at 1 to account (roughly) for all the
		; time spent computing since the pause started.  Total
		; accuracy isn't necessary here.

		; Stop timing and return A nonzero if the pause is longer
		; than the pause between "key still down" codes.
		
timepause:	mov	R1,#1
		mov	R2,#0

tploop:		nop                     		        ; 1  ..1
		nop						; 1  ..2
		nop						; 1  ..3
		nop						; 1  ..4
		nop						; 1  ..5
		nop						; 1  ..6
		nop						; 1  ..7
		nop						; 1  ..8
		nop						; 1  ..9
                acall   dolcd                                   ; 26 ..35

          	mov	a,R1					; 1  ..36
		add	a,#2					; 1  ..37
		mov	R1,a					; 1  ..38
		mov	a,R2					; 1  ..39
		addc	a,#0					; 1  ..40
		mov	R2,a					; 1  ..41
		nop						; 1  ..42
		cjne	a,#0x12,loopbot				; 2  ..44
		ret
loopbot:	jb	P3.5,tploop				; 2  ..46
		clr	a
		ret

                ; Subroutine to send "." followed by carriage return to
                ; the serial port

senddot:	mov	a,#'.'
		acall	putchar
		mov	a,#13
		sjmp	putchar

                ; Subroutine to send a 2-digit hex number to the serial port

printhex2:      mov     R0,a
                swap    a
                acall   prtdig
                mov     a,R0
prtdig:         anl     a,#0fh
                cjne    a,#10,prt2
prt2:           jc      notalpha
                add     a,#7
notalpha:       add     a,#'0'

                ; Subroutine to send a character to the serial port

putchar:        jnb     SCON.1,pc_busy
                mov     SBUF,A
                clr     SCON.1
                ret
pc_busy:	push	ACC
                acall	dolcd
                pop	ACC
		sjmp	putchar

                ; This routine takes 24 clock cycles regardless of what it does.
                ; If called at least twice per serial port minimum character time,
                ; it will feed the characters to the LCD.

                ; LCD commands are 0, command code.  This means that command 0
                ; can't be used.  No big deal (it's the write right-to-left command.)

dolcd:          jnb     SCON.0,nochar                          ; 2
                mov     a,SBUF                                 ; 1  .. 3
                clr     SCON.0                                 ; 1  .. 4
                cjne	a,#0xff,notff                          ; 2  .. 6
                sjmp    label9                                 ; 2  .. 8

notff:          mov	code,a				       ; 1  .. 7
                jnz     label12                                ; 2  .. 9
                clr     cdflag                                 ; 1  .. 10
                sjmp    busy                                   ; 2  .. 12

nochar:         mov	a,code                                 ; 1  .. 3
                jnz	tosend                                 ; 2  .. 5
                nop                                            ; 1  .. 6
                nop                                            ; 1  .. 7
                nop                                            ; 1  .. 8
label9:         nop                                            ; 1  .. 9
                nop                                            ; 1  .. 10
                nop                                            ; 1  .. 11
label12:        nop                                            ; 1  .. 12
busy:           nop                                            ; 1  .. 13
                nop                                            ; 1  .. 14
label15:        nop                                            ; 1  .. 15
                nop                                            ; 1  .. 16
                nop                                            ; 1  .. 17
                nop                                            ; 1  .. 18
                nop                                            ; 1  .. 19
                nop                                            ; 1  .. 20
                nop                                            ; 1  .. 21
                nop                                            ; 1  .. 22
                ret                                            ; 2  .. 24

tosend:		nop	                ; Command              ; 1  .. 6
		nop                     ; Tristate bus         ; 1  .. 7
		nop 	                ; Read                 ; 1  .. 8
		nop                     ; Assert E             ; 1  .. 9
		nop                     ; Deassert E           ; 1  .. 10
                nop
		nop                     ; check busy bit       ; 2  .. 12

                mov     C,cdflag                               ; 1  .. 14
                mov     P3.2,C          ; Command/data         ; 2  .. 16
                clr     P3.3            ; Write                ; 1  .. 17
                mov     P1,a            ; Data out             ; 1  .. 18
                setb    P3.4            ; Assert E             ; 1  .. 19
                clr     P3.4            ; Deassert E           ; 1  .. 20
                mov     code,#0                                ; 1  .. 21
                setb    cdflag                                 ; 1  .. 22
                ret                                            ; 2  .. 24
