#include "jukebox.h"

struct album album_data[MAX_ALBUMS];

int albums = 0;
int real_albums = 0;
int radio_record_album;

struct track track_data[MAX_TRACKS];

int tracks = 0;
int real_tracks = 0;

int current_track;

init_songdb()
{
   int current_album;
   char buf[1024];
   FILE *fp = fopen("musiclist","r");
   if(!fp) {
      fprintf(stderr,"Can't open music list\n");
      exit(999);
   }
   while(fgets(buf,sizeof(buf),fp)) {
      char *p;
      p = buf + strlen(buf)-1;
      while(p>=buf && isspace(*p)) *p-- = 0;
      if(*buf) {
         if(*buf==':') {
            char *artist = index(buf,'|');
            if(artist) *artist++ = 0;
            current_album = albums++;
            album_data[current_album].title = strdup(buf+1);
            album_data[current_album].artist = artist ? strdup(artist) : 0;
            album_data[current_album].first_track = tracks;
            //printf("Album #%d: %s\n",current_album,album_data[current_album].title);
         } else {
            char *nblk = index(buf,'|');
            char *title = 0;
            char *artist = 0;
            if(nblk) {
               //printf("found nblk: %s\n",nblk);
               *nblk++ = 0;
               title = index(nblk,'|');
               if(title) {
                  //printf("found title: %s\n",title);
                  *title++ = 0;
                  artist = index(title,'|');
                  if(artist) {
                     //printf("found artist: %s\n",artist);
                     *artist++ = 0;
                  }
               }
            }
            
            track_data[tracks].name = strdup(buf);
            track_data[tracks].title = title ? strdup(title) : track_data[tracks].name;
            track_data[tracks].artist = artist ? strdup(artist) : 0;
            track_data[tracks].album = current_album;
            track_data[tracks].length = atoi(nblk);
            track_data[tracks].fave_flags = 0;
//            printf("Track: %s\nlength=%d\ntitle=%s\n",track_data[tracks].name,
//               track_data[tracks].length,track_data[tracks].title);
            tracks++;
         }
      }
   } 
   fclose(fp);

   load_favourites();

   current_track = 0;
   real_tracks = tracks;
   real_albums = albums;
   find_recordings();
}

load_favourites()
{
   FILE *fp = fopen("favourites","r");
   char *p;
   int flags;
   int i;
   char buf[1024];
   if(!fp) {
      fprintf(stderr,"No favourite track list\n");
      return;
   }
   while(fgets(buf,sizeof(buf),fp)) {
      p = buf + strlen(buf)-1;
      while(p>=buf && isspace(*p)) *p-- = 0;
      sscanf(buf,"%x",&flags);
      for(i=0;i<tracks;i++) if(!strcmp(buf+4,track_data[i].name)) {
         track_data[i].fave_flags = flags;
         break;
      }
      if(i==tracks) {
         printf("bogus track in favourite list: %s\n",buf+4);
      }
   }
   fclose(fp);

   fp = fopen("favourites.clean","w");
   if(!fp) {
      fprintf(stderr,"Can't write cleaned-up favourites list");
      return;
   }
   for(i=0;i<tracks;i++) if(track_data[i].fave_flags) {
      fprintf(fp,"%03x %s\n",track_data[i].fave_flags,track_data[i].name);
   }
   fclose(fp);
}

int find_count;

int find_tracks(char *patbuf)
{
   int i;
   int nfs=-1;
   int found_song = -1;
   find_count = 0;
   //printf("find_tracks entered...\n");
   for(i=0;i<tracks;i++) {
      if(patmatch(1,patbuf,track_data[i].title)) {
         track_data[i].found = 1;
         find_count++;
         if(nfs<0) nfs=i;
      } else {
         track_data[i].found = 0;
      }
   }
   if(found_song<0 || !track_data[found_song].found) found_song=nfs;
   return found_song;
}




find_recordings()
{
   char buf[1024];
   FILE *fp;
   int i;
   for(i=real_tracks;i<tracks;i++) {
      free(track_data[i].name);
      free(track_data[i].title);
   }
   tracks = real_tracks;
   albums = real_albums;

   radio_record_album = real_albums;
   album_data[radio_record_album].title = "Radio Recordings";
   album_data[radio_record_album].artist = "";
   album_data[radio_record_album].first_track = real_tracks;

   sprintf(buf,"ls %s/record-????????-?????? %s/record-????????-??????.raw %s/record-????????-??????.mp3",
      RECFILE_PATH,RECFILE_PATH,COMPRESSED_RECFILE_PATH);
   fp = popen(buf,"r");
   if(!fp) {
      perror("popen");
      fprintf(stderr,"Could not find recordings\n");
      return;
   }
   while(fgets(buf,sizeof(buf),fp)) {
      char *p = rindex(buf,'/');
      char *q, *r;
      p = p ? p+1 : buf;
      p[22] = 0;
      //printf("p=\"%s\", rt=%d, t=%d\n",p,real_tracks,tracks);
      for(i=real_tracks;i<tracks;i++) if(!strcmp(track_data[i].name,p)) break;
      if(i<tracks) continue;
      track_data[tracks].name = strdup(p);
      track_data[tracks].album = radio_record_album;
      track_data[tracks].artist = "Radio Recording";
      track_data[tracks].fave_flags = 0;
      p = p+7;
      //printf("magic: p=%s\n",p);
      q = "..../../.. !..:..:..";
      r = buf;
      while(*q) {
         if(*q=='.') *r++ = *p++;
         else if(*q=='!') p++;
         else *r++ = *q;
         q++;
      }
      *r = 0;
      //printf("%s\n",buf);
      track_data[tracks].title = strdup(buf);
      albums = real_albums + 1;
      tracks++;
   }
   fclose(fp);
   if(current_track >= tracks) current_track = tracks-1;
//   printf("recordings scanned - tracks=%d, real_tracks=%d, albums=%d, real_albums=%d\n",
//      tracks,real_tracks,albums,real_albums);
}
